"""A setuptools based setup module for DiviK algorithm."""

from glob import glob
import os
from setuptools import setup, find_packages, Extension
import sys
import numpy

__version__ = '2.5.9b'

LINUX_OPTS = {
    'extra_link_args': [
        '-fopenmp',
    ],
    'extra_compile_args': [
        '-fopenmp',
        '-Wno-strict-prototypes',
        '-Wno-maybe-uninitialized',
        '-O3',
        '-std=c99',
    ],
}
OSX_OPTS = {
    'extra_link_args': [
        '-L/usr/local/opt/libomp/lib',
        '-fopenmp',
    ],
    'extra_compile_args': [
        '-Wno-strict-prototypes',
        '-Wno-uninitialized',
        '-fopenmp',
        '-O3',
    ],
}
WINDOWS_OPTS = {
    'extra_compile_args': [
        '/Ox',
    ]
}


if os.name == 'posix':
    if sys.platform.startswith('darwin'):
        OPTS = OSX_OPTS
    else:
        OPTS = LINUX_OPTS
else:
    OPTS = WINDOWS_OPTS

with open('README.md') as infile:
    readme = infile.read()

setup(
    name='divik',
    version=__version__,
    description='Divisive iK-means algorithm implementation',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/gmrukwa/divik',
    author='Grzegorz Mrukwa',
    author_email='g.mrukwa@gmail.com',
    classifiers=[
        # based on https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ],
    entry_points={
        'console_scripts': [
            'fit-clusters=divik._cli.fit_clusters:main',
            'dunn-divik=divik._cli.dunn_divik:main',
            'divik=divik._cli.divik:main',
            'inspect=divik._cli.inspect:main',
            'kmeans=divik._cli.auto_kmeans:main',
            'linkage=divik._cli.linkage:main',
            'spectral=divik._cli.spectral:main',
            'visualize=divik._cli.visualize:main'
        ],
    },
    packages=find_packages(exclude=['test']),
    # @gmrukwa: https://packaging.python.org/discussions/install-requires-vs-requirements/
    install_requires=[
        'dash==0.34.0',
        'dash-html-components==0.13.4',
        'dash-core-components==0.42.0',
        'dash-table==3.1.11',
        'dask[dataframe]>=2.14.0',
        'dask-distance>=0.2.0',
        'h5py>=2.8.0',
        'kneed>=0.5.1',
        'numpy>=0.12.1',
        'pandas>=0.20.3',
        'scipy>=0.19.1',
        'scikit-image>=0.14.1',
        'scikit-learn>=0.19.1',
        'tqdm>=4.11.2',
    ],
    setup_requires=[
        'numpy>=0.12.1',
    ],
    extras_require={
        'gin': [
            "absl-py",
            "gin-config",
        ],
    },
    python_requires='>=3.6',
    package_data={
    },
    ext_modules=[
        Extension('gamred_native',
                  sources=glob('gamred_native/*.c'),
                  include_dirs=['gamred_native', numpy.get_include()],
                  define_macros=[
                      ('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION'),
                  ],
                  **OPTS),
    ],
)



#!/usr/bin/env python
import logging

from divik.core import build
from divik.cluster import DiviK, DunnSearch, GAPSearch, KMeans
import divik._cli._utils as sc

from divik._cli.divik import save


def _fast_kmeans(**config):
    distance = config.get('distance', 'correlation')
    normalize_rows = config.get('normalize_rows', None)
    if normalize_rows is None:
        normalize_rows = distance == 'correlation'
    single_kmeans = KMeans(
        n_clusters=2,
        distance=distance,
        init='percentile',
        percentile=config.get('distance_percentile', 99.0),
        max_iter=config.get('max_iter', 10),
        normalize_rows=normalize_rows,
    )
    kmeans = GAPSearch(
        single_kmeans,
        max_clusters=2,
        n_jobs=config.get('n_jobs', None),
        seed=config.get('random_seed', 0),
        n_trials=config.get('gap_trials', 10),
        sample_size=config.get('sample_size', 10000),
        verbose=config.get('verbose', False),
    )
    return kmeans


def _full_kmeans(**config):
    distance = config.get('distance', 'correlation')
    normalize_rows = config.get('normalize_rows', None)
    single_kmeans = KMeans(
        n_clusters=2,
        distance=distance,
        init='percentile',
        percentile=config.get('distance_percentile', 99.0),
        max_iter=config.get('max_iter', 100),
        normalize_rows=normalize_rows,
    )
    kmeans = DunnSearch(
        single_kmeans,
        max_clusters=config.get('k_max', 10),
        n_jobs=config.get('n_jobs', None),
        verbose=config.get('verbose', False),
    )
    return kmeans


def main():
    data, config, destination, xy = sc.initialize()
    logging.info('Workspace initialized.')
    logging.info('Scenario configuration: {0}'.format(config))
    fast = _fast_kmeans(**config)
    full = _full_kmeans(**config)
    divik_config = {'kmeans': full, 'fast_kmeans': fast, **config}
    divik = build(DiviK, **divik_config)
    logging.info("Launching experiment.")
    try:
        divik.fit(data)
    except Exception as ex:
        logging.error("Failed with exception.")
        logging.error(repr(ex))
        raise
    save(data, divik, destination, xy)


if __name__ == '__main__':
    main()



from functools import partial
from typing import Callable, List, Tuple

import numpy as np
from skimage import feature as ft

from ._matlab_alike import n_quantiles


_DISCRETIZATION_LEVELS = 8
_quantile_thresholds = partial(n_quantiles,
                               N=_DISCRETIZATION_LEVELS - 1)
_greycomatrix_backend = partial(ft.greycomatrix,
                                distances=[1, np.sqrt(2), 1, np.sqrt(2)],
                                angles=np.radians([0., -45., -90., -135.]),
                                symmetric=False,
                                normed=False)


def _greycomatrix(discrete_image: np.ndarray, mask: np.ndarray) -> np.ndarray:
    all_allowed = np.all(mask)
    if all_allowed:
        levels = _DISCRETIZATION_LEVELS
    else:
        levels = _DISCRETIZATION_LEVELS + 1
        discrete_image[~mask] = np.max(discrete_image[mask]) + 1
    gcm = _greycomatrix_backend(discrete_image, levels=levels)
    if not all_allowed:
        gcm = gcm[:-1, :-1, :, :]
    return np.dstack([gcm[:, :, i, i] for i in range(gcm.shape[2])])


def _ignorance_mask(image: np.ndarray, ignored: List) -> np.ndarray:
    mask = np.ones(image.shape, dtype=bool)
    for value in ignored:
        mask = np.logical_and(mask, image != value)
    return mask


def _discretize(image: np.ndarray, mask: np.ndarray) -> np.ndarray:
    image = image.astype(float)
    image /= np.max(image[mask])
    thresholds = np.hstack([_quantile_thresholds(image[mask].ravel()), (1.,)])
    discrete = np.zeros(image.shape, dtype=np.uint8)
    for i in range(len(thresholds)-1, -1, -1):
        pixels_of_interest = np.logical_and(image <= thresholds[i], mask)
        discrete[pixels_of_interest] = i
    return discrete


_split_directions = partial(np.rollaxis, axis=2)
_EXPECTED_BLOCK_SHAPE = 2 * (_DISCRETIZATION_LEVELS / 2,)


def _block_structness(greycomatrix_block: np.ndarray) -> float:
    if greycomatrix_block.shape != _EXPECTED_BLOCK_SHAPE:
        raise ValueError('Expected shape {0}, got {1}.'
                         .format(_EXPECTED_BLOCK_SHAPE, greycomatrix_block.shape))
    return np.sum(greycomatrix_block[:3, :3]) \
        + np.sum(greycomatrix_block[:2, :2]) \
        + 2. * greycomatrix_block[0, 0]


_BlockSelector = Callable[[np.ndarray], np.ndarray]


def _darkness(greycomatrix: np.ndarray) -> np.ndarray:
    size = int(_DISCRETIZATION_LEVELS / 2)
    return greycomatrix[:size, :size]


def _lightness(greycomatrix: np.ndarray) -> np.ndarray:
    size = int(_DISCRETIZATION_LEVELS / 2 + 1)
    return greycomatrix[:-size:-1, :-size:-1]


def _structness_of(selector: _BlockSelector, directions: np.ndarray) -> float:
    return float(np.sum(
        _block_structness(selector(direction))
        for direction in directions
    ))


def structness(image: np.ndarray, ignored: List=None) -> Tuple[float, float]:
    if ignored is None:
        ignored = []
    if len(image.shape) != 2:
        raise ValueError('Expected 2D image, got {0}D.'.format(len(image.shape)))
    mask = _ignorance_mask(image, ignored)
    discrete = _discretize(image, mask)
    greycomatrix = _greycomatrix(discrete, mask)
    directions = _split_directions(greycomatrix)
    structness_of_darkness = _structness_of(_darkness, directions)
    structness_of_lightness = _structness_of(_lightness, directions)
    return structness_of_darkness, structness_of_lightness



from contextlib import contextmanager
import inspect
# RawArray exists, but PyCharm goes crazy
# noinspection PyUnresolvedReferences
from multiprocessing import Pool, RawArray

import numpy as np
from skimage.color import label2rgb

from ._types import Data


def normalize_rows(data: Data) -> Data:
    normalized = data - data.mean(axis=1)[:, np.newaxis]
    norms = np.sum(np.abs(normalized) ** 2, axis=-1, keepdims=True)**(1./2)
    normalized /= norms
    return normalized


def visualize(label, xy, shape=None):
    x, y = xy.T
    if shape is None:
        shape = np.max(y) + 1, np.max(x) + 1
    y = y.max() - y
    label = label - label.min() + 1
    label_map = np.zeros(shape, dtype=int)
    label_map[y, x] = label
    image = label2rgb(label_map, bg_label=0)
    return image


@contextmanager
def context_if(condition, context, *args, **kwargs):
    if condition:
        with context(*args, **kwargs) as c:
            yield c
    else:
        yield None


def build(klass, **kwargs):
    """Build instance of klass using matching kwargs"""
    known_param_names = list(inspect.signature(klass).parameters.keys())
    known_params = {n: kwargs[n] for n in known_param_names if n in kwargs}
    return klass(**known_params)



import json
import logging
import os
import pickle

import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline

from divik.core import configurable, DivikResult


_SAVERS = set()


def saver(fn):
    _SAVERS.add(fn)


def save(model, destination, **kwargs):
    for save_fn in _SAVERS:
        save_fn(model, destination, **kwargs)


@saver
@configurable(whitelist=['enabled'])
def save_pickle(model, destination, enabled=True, **kwargs):
    if not enabled:
        return
    logging.info('Saving model pickle.')
    with open(os.path.join(destination, 'model.pkl'), 'wb') as pkl:
        pickle.dump(model, pkl)


@saver
def save_divik(model, destination, **kwargs):
    if not hasattr(model, 'result_'):
        return
    if not isinstance(model.result_, DivikResult):
        logging.info("Skipping DiviK details save. Cause: result is None")
        return
    from .divik import make_merged, save_merged
    logging.info('Saving DiviK details.')
    logging.info('Saving DivikResult pickle.')
    with open(os.path.join(destination, 'result.pkl'), 'wb') as pkl:
        pickle.dump(model.result_, pkl)
    logging.info("Saving DiviK partitions.")
    merged = make_merged(model.result_).astype(np.int64)
    assert merged.shape[0] == model.result_.clustering.labels_.size
    xy = kwargs.get('xy', None)
    save_merged(destination, merged, xy)


@saver
def save_summary(model, destination, **kwargs):
    if not hasattr(model, 'labels_') or \
            not hasattr(model, 'n_clusters_') or \
            not hasattr(model, 'depth_'):
        return
    logging.info("Saving JSON summary.")
    with open(os.path.join(destination, 'summary.json'), 'w') as smr:
        json.dump({
            "depth": model.depth_,
            "number_of_clusters": model.n_clusters_,
            "mean_cluster_size": model.labels_.size / float(model.n_clusters_)
        }, smr)


@saver
def save_labels(model, destination, **kwargs):
    if not hasattr(model, 'labels_'):
        return
    logging.info("Saving final partition.")
    np.save(os.path.join(destination, 'final_partition.npy'), model.labels_)
    np.savetxt(os.path.join(destination, 'final_partition.csv'), model.labels_,
               delimiter=', ', fmt='%i')


@saver
def save_centroids(model, destination, **kwargs):
    if not hasattr(model, 'centroids_'):
        return
    logging.info("Saving centroids.")
    np.save(os.path.join(destination, 'centroids.npy'), model.centroids_)
    np.savetxt(os.path.join(destination, 'centroids.csv'), model.centroids_,
               delimiter=', ')


@saver
def save_filters(model, destination, **kwargs):
    if not hasattr(model, 'filters_'):
        return
    logging.info("Saving filters.")
    np.save(os.path.join(destination, 'filters.npy'), model.filters_)
    np.savetxt(os.path.join(destination, 'filters.csv'), model.filters_,
               delimiter=', ', fmt='%i')


@saver
def save_cluster_paths(model, destination, **kwargs):
    if not hasattr(model, 'reverse_paths_'):
        return
    rev = ['_'.join(map(str, p)) for p in model.reverse_paths_]
    pd.DataFrame({
        'path': rev,
        'cluster_number': list(model.reverse_paths_.values())
    }).to_csv(os.path.join(destination, 'paths.csv'))

@saver
def save_pipeline(model, destination, **kwargs):
    if not isinstance(model, Pipeline):
        return
    feature_selector = model[:-1]
    clustering = model[-1]
    if isinstance(clustering, Pipeline):
        logging.info('Saving pre-extractor pickle.')
        with open(os.path.join(destination, 'feature_pre_extractor.pkl'), 'wb') as pkl:
            pickle.dump(feature_selector, pkl)
        return save(clustering, destination, **kwargs)
    logging.info('Saving model pickle.')
    with open(os.path.join(destination, 'feature_selector.pkl'), 'wb') as pkl:
        pickle.dump(feature_selector, pkl)
    save(clustering, destination, **kwargs)
    if not os.path.exists(os.path.join(destination, 'summary.json')):
        logging.info("Saving JSON summary.")
        with open(os.path.join(destination, 'summary.json'), 'w') as smr:
            json.dump({
                "depth": 1,
                "number_of_clusters": int(clustering.n_clusters_),
                "mean_cluster_size": \
                    clustering.labels_.size / float(clustering.n_clusters_)
            }, smr)
    if not os.path.exists(os.path.join(destination, 'final_partition.npy')):
        logging.info("Saving final partition.")
        np.save(os.path.join(destination, 'final_partition.npy'), clustering.labels_)
        np.savetxt(os.path.join(destination, 'final_partition.csv'), clustering.labels_,
                delimiter=', ', fmt='%i')
    if not os.path.exists(os.path.join(destination, 'partition-0.png')):
        from .divik import save_merged
        save_merged(
            destination,
            clustering.labels_.reshape(-1, 1),
            xy=kwargs.get('xy', None)
        )



from itertools import chain
import glob
import logging
import os
from functools import partial
from typing import List

import h5py
import numpy as np
from scipy import io as scio

import divik.core as u


def _load_mat_with(path: str, backend=scio.loadmat, ignore='__') -> np.ndarray:
    data = backend(path)
    logging.debug('Data file opened successfully.')
    key = [key for key in list(data.keys()) if not key.startswith(ignore)]
    logging.debug('Found variables: {0}'.format(key))
    if len(key) != 1:
        raise ValueError('There should be a single variable inside MAT-file: '
                         + path + '\nWere: ' + str(key))
    logging.debug('Selecting variable: {0}'.format(key[0]))
    selected = data[key[0]]
    logging.debug('Loaded variable from file.')
    contignuous = np.array(selected, dtype=float)
    logging.debug('Converted to contignuous.')
    return contignuous


def _load_mat(path: str) -> np.ndarray:
    logging.debug('Loading MAT-file: ' + path)
    try:
        logging.debug('Trying out legacy MAT-file loader.')
        return _load_mat_with(path, backend=scio.loadmat, ignore='__')
    except NotImplementedError:  # v7.3 MATLAB HDF5 MAT-File
        logging.debug('Legacy MAT-file loader failed, restarting with HDF5 loader.')
        hdf5 = partial(h5py.File, mode='r')
        return _load_mat_with(path, backend=hdf5, ignore='#').T


def load_data(path: str) -> u.Data:
    logging.info("Loading data: " + path)
    normalized = path.lower()
    if normalized.endswith('.csv'):
        loader = partial(np.loadtxt, delimiter=',')
    elif normalized.endswith('.txt'):
        loader = np.loadtxt
    elif normalized.endswith('.npy'):
        loader = np.load
    elif normalized.endswith('.mat'):
        loader = _load_mat
    else:
        message = 'Unsupported data format: ' + os.path.splitext(path)[1]
        logging.error(message)
        raise IOError(message)
    return loader(path)


DIVIK_RESULT_FNAME = 'result.pkl'


def _result_path_patterns(slug: str) -> List[str]:
    slug_pattern = '*{0}*'.format(slug)
    direct = os.path.join(slug_pattern, DIVIK_RESULT_FNAME)
    prefixed = os.path.join('**', slug_pattern, DIVIK_RESULT_FNAME)
    suffixed = os.path.join(slug_pattern, '**', DIVIK_RESULT_FNAME)
    bothfixed = os.path.join('**', slug_pattern, '**', DIVIK_RESULT_FNAME)
    return list((direct, prefixed, suffixed, bothfixed))


def _find_possible_directories(patterns: List[str]) -> List[str]:
    possible_locations = chain.from_iterable(
        glob.glob(pattern, recursive=True) for pattern in patterns)
    possible_paths = list({
        os.path.split(fname)[0] for fname in possible_locations
    })
    return possible_paths


def as_divik_result_path(path_or_slug: str):
    possible_location = os.path.join(path_or_slug, DIVIK_RESULT_FNAME)
    if os.path.exists(possible_location):
        return path_or_slug
    patterns = _result_path_patterns(path_or_slug)
    possible_paths = _find_possible_directories(patterns)
    if not possible_paths:
        raise FileNotFoundError(path_or_slug)
    if len(possible_paths) > 1:
        msg = 'Multiple possible result directories: {0}. Selecting {1}.'
        logging.warning(msg.format(possible_paths, possible_paths[0]))
    return possible_paths[0]



from functools import partial
import sys
from typing import Tuple

import numpy as np
import pandas as pd
import tqdm
from scipy.spatial import distance as dist
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from divik import _summary as summary, feature_selection as fs
from divik.core import (
    configurable,
    context_if,
    DivikResult,
    normalize_rows,
    maybe_pool,
)
from ._backend import divik
from ._report import DivikReporter


@configurable
class DiviK(BaseEstimator, ClusterMixin, TransformerMixin):
    """DiviK clustering

    Parameters
    ----------
    kmeans: AutoKMeans
        A self-tuning KMeans estimator for the purpose of clustering

    fast_kmeans: GAPSearch, optional, default: None
        A self-tuning KMeans estimator for the purpose of stop condition
        check. If None, the `kmeans` parameter is assumed to be the
        `GAPSearch` instance.

    distance: str, optional, default: 'correlation'
        The distance metric between points, centroids and for GAP index
        estimation. One of the distances supported by scipy package.

    minimal_size: int, optional, default: None
        The minimum size of the region (the number of observations) to be
        considered for any further divisions. When left None, defaults to
        0.1% of the training dataset size.

    rejection_size: int, optional, default: None
        Size under which split will be rejected - if a cluster appears in the
        split that is below rejection_size, the split is considered improper
        and discarded. This may be useful for some domains (like there is no
        justification for a 3-cells cluster in biological data). By default,
        no segmentation is discarded, as careful post-processing provides the
        same advantage.

    rejection_percentage: float, optional, default: None
        An alternative to ``rejection_size``, with the same behavior, but this
        parameter is related to the training data size percentage. By default,
        no segmentation is discarded.

    minimal_features_percentage: float, optional, default: 0.01
        The minimal percentage of features that must be preserved after
        GMM-based feature selection. By default at least 1% of features is
        preserved in the filtration process.

    features_percentage: float, optional, default: 0.05
        The target percentage of features that are used by fallback percentage
        filter for 'outlier' filter.

    normalize_rows: bool, optional, default: None
        Whether to normalize each row of the data to the norm of 1. By default,
        it normalizes rows for correlation metric, does no normalization
        otherwise.

    use_logfilters: bool, optional, default: False
        Whether to compute logarithm of feature characteristic instead of the
        characteristic itself. This may improve feature filtering performance,
        depending on the distribution of features, however all the
        characteristics (mean, variance) have to be positive for that -
        filtering will fail otherwise. This is useful for specific cases in
        biology where the distribution of data may actually require this option
        for any efficient filtering.

    filter_type: {'gmm', 'outlier', 'auto', 'none'}, default: 'gmm'
        - 'gmm' - usual Gaussian Mixture Model-based filtering, useful for high
        dimensional cases
        - 'outlier' - robust outlier detection-based filtering, useful for low
        dimensional cases. In the case of no outliers, percentage-based
        filtering is applied.
        - 'auto' - automatically selects between 'gmm' and 'outlier' based on
        the dimensionality. When more than 250 features are present, 'gmm'
        is chosen.
        - 'none' - feature selection is disabled

    n_jobs: int, optional, default: None
        The number of jobs to use for the computation. This works by computing
        each of the GAP index evaluations in parallel and by making predictions
        in parallel.

    verbose: bool, optional, default: False
        Whether to report the progress of the computations.

    Attributes
    ----------

    result_: divik.DivikResult
        Hierarchical structure describing all the consecutive segmentations.

    labels_:
        Labels of each point

    centroids_: array, [n_clusters, n_features]
        Coordinates of cluster centers. If the algorithm stops before fully
        converging, these will not be consistent with ``labels_``. Also, the
        distance between points and respective centroids must be captured
        in appropriate features subspace. This is realized by the ``transform``
        method.

    filters_: array, [n_clusters, n_features]
        Filters that were applied to the feature space on the level that was
        the final segmentation for a subset.

    depth_: int
        The number of hierarchy levels in the segmentation.

    n_clusters_: int
        The final number of clusters in the segmentation, on the tree leaf
        level.

    paths_: Dict[int, Tuple[int]]
        Describes how the cluster number corresponds to the path in the tree.
        Element of the tuple indicates the sub-segment number on each tree
        level.

    reverse_paths_: Dict[Tuple[int], int]
        Describes how the path in the tree corresponds to the cluster number.
        For more details see ``paths_``.

    Examples
    --------

    >>> from divik.cluster import DunnDiviK
    >>> from sklearn.datasets import make_blobs
    >>> X, _ = make_blobs(n_samples=200, n_features=100, centers=20,
    ...                   random_state=42)
    >>> divik = DiviK(distance='euclidean').fit(X)
    >>> divik.labels_
    array([1, 1, 1, 0, ..., 0, 0], dtype=int32)
    >>> divik.predict([[0, ..., 0], [12, ..., 3]])
    array([1, 0], dtype=int32)
    >>> divik.cluster_centers_
    array([[10., ...,  2.],
           ...,
           [ 1, ...,  2.]])

    """
    def __init__(self,
                 kmeans,
                 fast_kmeans=None,
                 distance: str = 'correlation',
                 minimal_size: int = None,
                 rejection_size: int = None,
                 rejection_percentage: float = None,
                 minimal_features_percentage: float = .01,
                 features_percentage: float = 0.05,
                 normalize_rows: bool = None,
                 use_logfilters: bool = False,
                 filter_type='gmm',
                 n_jobs: int = None,
                 verbose: bool = False):
        self.kmeans = kmeans
        self.fast_kmeans = fast_kmeans
        self.distance = distance
        self.minimal_size = minimal_size
        self.rejection_size = rejection_size
        self.rejection_percentage = rejection_percentage
        self.minimal_features_percentage = minimal_features_percentage
        self.features_percentage = features_percentage
        self.normalize_rows = normalize_rows
        self.use_logfilters = use_logfilters
        self.filter_type = filter_type
        self.n_jobs = n_jobs
        self.verbose = verbose
        self._validate_arguments()

    def _validate_arguments(self):
        if self.minimal_features_percentage < 0 \
                or self.minimal_features_percentage > 1:
            raise ValueError('minimal_features_percentage must be in range'
                             ' [0, 1]')
        if self.features_percentage < 0 or self.features_percentage > 1:
            raise ValueError('features_percentage must be in range [0, 1]')
        if self.features_percentage < self.minimal_features_percentage:
            raise ValueError('features_percentage must be higher than or equal'
                             ' to minimal_features_percentage')
        if self.filter_type not in ['gmm', 'outlier', 'auto', 'none']:
            raise ValueError(
                "filter_type must be in ['gmm', 'outlier', 'auto', 'none']")

    def fit(self, X, y=None):
        """Compute DiviK clustering.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.
        y : Ignored
            not used, present here for API consistency by convention.
        """
        if np.isnan(X).any():
            raise ValueError("NaN values are not supported.")

        with context_if(self.verbose, tqdm.tqdm, total=X.shape[0],
                        file=sys.stdout, smoothing=0) as progress:
            self.result_ = self._divik(X, progress)

        if self.result_ is None:
            self.labels_ = np.zeros((X.shape[0],), dtype=int)
            self.paths_ = {0: (0,)}
        else:
            self.labels_, self.paths_ = summary.merged_partition(
                self.result_, return_paths=True)

        self.reverse_paths_ = {
            value: key for key, value in self.paths_.items()}

        if self.result_ is None:
            self.filters_ = np.ones([1, X.shape[1]], dtype=bool)
        else:
            self.filters_ = np.array(
                [self._get_filter(path) for path in self.reverse_paths_],
                dtype=bool)
        self.centroids_ = pd.DataFrame(X).groupby(self.labels_, sort=True)\
            .mean().values
        self.depth_ = summary.depth(self.result_)
        self.n_clusters_ = summary.total_number_of_clusters(self.result_)

        return self

    def _get_rejection_size(self, X):
        rejection_size = 0
        if self.rejection_size is not None:
            rejection_size = max(rejection_size, self.rejection_size)
        if self.rejection_percentage is not None:
            rejection_size = max(
                rejection_size, int(self.rejection_percentage * X.shape[0]))
        return rejection_size

    def _get_filter(self, path):
        """This method extracts features filter used for each centroid"""
        result = self.result_
        for item in path[:-1]:
            result = result.subregions[item]
        return result.feature_selector.selected_

    def _needs_normalization(self):
        if self.normalize_rows is None:
            return self.distance == 'correlation'
        return self.normalize_rows

    def _feature_selector(self, n_features):
        return fs.make_specialized_selector(
            self.filter_type, n_features, use_log=self.use_logfilters,
            min_features_rate=self.minimal_features_percentage,
            p=self.features_percentage
        )

    def _divik(self, X, progress):
        full = self.kmeans
        fast = self.fast_kmeans
        if fast is None:
            warn_const = full.kmeans.normalize_rows
        else:
            warn_const = fast.kmeans.normalize_rows or full.kmeans.normalize_rows
        report = DivikReporter(progress, warn_const=warn_const)
        select_all = np.ones(shape=(X.shape[0],), dtype=bool)
        minimal_size = int(X.shape[0] * 0.001) if self.minimal_size is None \
            else self.minimal_size
        rejection_size = self._get_rejection_size(X)
        return divik(
            X, selection=select_all, kmeans=full, fast_kmeans=fast,
            feature_selector=self._feature_selector(X.shape[1]),
            minimal_size=minimal_size, rejection_size=rejection_size,
            report=report)

    def fit_predict(self, X, y=None):
        """Compute cluster centers and predict cluster index for each sample.

        Convenience method; equivalent to calling fit(X) followed by
        predict(X).

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        y : Ignored
            not used, present here for API consistency by convention.

        Returns
        -------

        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        return self.fit(X).labels_

    def transform(self, X):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        Returns
        -------

        X_new : array, shape [n_samples, self.n_clusters_]
            X transformed in the new space.
        """
        check_is_fitted(self)
        if self._needs_normalization():
            X = normalize_rows(X)
        distances = np.hstack([
            dist.cdist(
                X[:, selector], centroid[np.newaxis, selector], self.distance)
            for selector, centroid in zip(self.filters_, self.centroids_)
        ])
        return distances

    def fit_transform(self, X, y=None, **fit_params):
        """Compute clustering and transform X to cluster-distance space.

        Equivalent to fit(X).transform(X), but more efficiently implemented.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        y : Ignored
            not used, present here for API consistency by convention.

        Returns
        -------

        X_new : array, shape [n_samples, self.n_clusters_]
            X transformed in the new space.
        """
        return self.fit(X).transform(X)

    def predict(self, X):
        """Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        Parameters
        ----------

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to predict.

        Returns
        -------

        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        check_is_fitted(self)
        if self._needs_normalization():
            X = normalize_rows(X)
        predict = partial(_predict_path, result=self.result_)
        with maybe_pool(self.n_jobs) as pool:
            paths = pool.map(predict, X)
        labels = [self.reverse_paths_[path] for path in paths]
        return np.array(labels, dtype=np.int32)


def _predict_path(observation: np.ndarray, result: DivikResult) -> Tuple[int]:
    path = []
    observation = observation[np.newaxis, :]
    division = result
    while division is not None:
        local_X = division.feature_selector.transform(observation)
        label = int(division.clustering.predict(local_X))
        path.append(label)
        division = division.subregions[label]
    path = tuple(path) if len(path) else (0,)
    return path



import numpy as np
from sklearn.base import BaseEstimator

from divik.core import configurable
from ._stat_selector_mixin import StatSelectorMixin


# noinspection PyAttributeOutsideInit
@configurable
class PercentageSelector(BaseEstimator, StatSelectorMixin):
    """Feature selector that removes / preserves top some percent of features

    This feature selection algorithm looks only at the features (X), not the
    desired outputs (y), and can thus be used for unsupervised learning.

    Parameters
    ----------
    stat: {'mean', 'var'}
        Kind of statistic to be computed out of the feature.

    use_log: bool, optional, default: False
        Whether to use the logarithm of feature characteristic instead of the
        characteristic itself. This may improve feature filtering performance,
        depending on the distribution of features, however all the
        characteristics (mean, variance) have to be positive for that -
        filtering will fail otherwise. This is useful for specific cases in
        biology where the distribution of data may actually require this option
        for any efficient filtering.

    keep_top: bool, optional, default: True
        When True, keeps features with highest value of the characteristic.

    p: float, optional, default: 0.2
        Rate of features to keep.

    Attributes
    ----------
    vals_: array, shape (n_features,)
        Computed characteristic of each feature.

    threshold_: float
        Value of the threshold used for filtering

    selected_: array, shape (n_features,)
        Vector of binary selections of the informative features.
    """
    def __init__(self, stat: str, use_log: bool = False, keep_top: bool = True,
                 p: float = 0.2):
        self.stat = stat
        self.use_log = use_log
        self.keep_top = keep_top
        self.p = p

    def fit(self, X, y=None):
        """Learn data-driven feature thresholds from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors from which to compute feature characteristic.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.

        Returns
        -------
        self
        """
        self.vals_ = self._to_characteristics(X)
        if self.keep_top:
            self.threshold_ = np.quantile(self.vals_, q=1 - self.p)
            self.selected_ = self.threshold_ <= self.vals_
        else:
            self.threshold_ = np.quantile(self.vals_, q=self.p)
            self.selected_ = self.threshold_ >= self.vals_
        return self
